#include <qpainter.h>
#include <qwidget.h>
#include <qimage.h>
#include <iostream.h>
#include <cmath>
#include <complex.h>
#include "gimshow.h"
#include "grayprocessingimage.h"

//! Constructor for the GImShow class
/*! Keeps the QImages and displays either the original image, the real, imag og magnitude part of the filter or of the filtered image. */
GImShow::GImShow(QWidget * parent=0, const char * name=0, WFlags f=0) :  QWidget(parent, name, f)
{
  origimage = new GrayProcessingImage;;
  filterreal = new GrayProcessingImage;
  filterimag = new GrayProcessingImage;
  filtermag = new GrayProcessingImage;
  resultreal = new GrayProcessingImage;
  resultimag = new GrayProcessingImage;
  resultmag = new GrayProcessingImage;
  viewimage = NULL;
  paint = new QPainter(this);
  repaint();
}

GImShow::~GImShow()
{
}

int GImShow::width()
{
  if(viewimage == NULL || viewimage->width() < 400)
    return 400;
  else
    return viewimage->width();
}


int GImShow::height()
{
  if(viewimage == NULL || viewimage->height() < 300)
    return 300;
  else
    return viewimage->height();
}


bool GImShow::loadImage(const char *fileName)
{
  //! Test to see if image->load accepts the filename, if so, draw the image
  if(origimage->load(fileName,0))
  {
    // if not 8-bit grayscale, convert
    if(origimage->depth() != 8)
      origimage->convertDepth(8,0);
    viewimage = origimage;
    repaint();
  }
  else
  {
    cout << "no image loaded" << endl;
    return 1;
  }
  return 0;
}
  
bool GImShow::saveImage(const char *filename)
{
  if(viewimage->save(filename,"BMP"))
    return 0;
  else
  {
    cout << "no image loaded" << endl;
    return 1;
  }
}

//! paintEvent function which controls the size of the entire application.
/*! This is clumsy implemented since there should be a separate resize event handler. The paintEvent function shoud simply handle the event generated repaints. */
void GImShow::paintEvent(QPaintEvent *e)
{
  if(viewimage != NULL)
  {    
    if(viewimage->width() < 400 && viewimage->height() < 300)
    {
      paint->setClipRect(e->rect());
      resize(400,300);
      paint->drawImage(200-(floor(viewimage->width()/2)),150-(floor(viewimage->height()/2)),*viewimage);  
    }
    else
    {
      paint->setClipRect(e->rect());
      resize(viewimage->width(),viewimage->height());
      paint->drawImage(0,0,*viewimage); 
    }
  }
  else //no image loaded, so just initialize with help text
  {
    resize(400,300);
    drawText(150,150,"No image loaded");
  }
}

void GImShow::slotViewOrigImage()
{
  viewimage = origimage;
  repaint();
}  

void GImShow::slotViewFilterReal()
{
  viewimage = filterreal;
  repaint();
}

void GImShow::slotViewFilterImag()
{
  viewimage = filterimag;
  repaint();
}

void GImShow::slotViewFilterMag()
{
  viewimage = filtermag;
  repaint();
}

void GImShow::slotViewResultReal()
{
  viewimage = resultreal;
  repaint();
}

void GImShow::slotViewResultImag()
{
  viewimage = resultimag;
  repaint();
}

void GImShow::slotViewResultMag()
{
  viewimage = resultmag;
  repaint();
}


//! The most important function which takes care of all the computation.
/*! The size of the filter is set dependant on the size of the Gabor kernel determined by the parameters set by the user. */
void GImShow::slotSetFilter(QString Ss, QString Ks, QString Ps, QString as, QString bs, QString x0s, QString y0s, QString u0s, QString v0s, QString thetas)
{

  // convert all filter parameters from QString to appropriate datatypes

  S = Ss.toInt(0,10);
  K = Ks.toDouble();
  P = Ps.toDouble();
  a = as.toDouble();
  b = bs.toDouble();
  theta = thetas.toDouble();
  x0 = x0s.toDouble();
  y0 = y0s.toDouble();
  u0 = u0s.toDouble();
  v0 = v0s.toDouble();

  // compute filter size from the parameters of the Gaussian
  /*
  if(a < b)
    size = int(ceil(2*(1/a)));
  else
    size = int(ceil(2*(1/b)));
  */

  // for now, we set the filter size manually

  size = S;

  // initialize filter

  filter = new complex<double>*[size];  
  for(int i=0;i<size;i++)
    filter[i] = new complex<double>[size];

  // initialize all max and min values

  maxfilterreal = 0; 
  minfilterreal = 0; 
  maxfilterimag = 0; 
  minfilterimag = 0; 
  maxfiltermag = 0; 
  minfiltermag = 0; 
  maxresultreal = 0; 
  minresultreal = 0; 
  maxresultimag = 0; 
  minresultimag = 0; 
  maxresultmag = 0; 
  minresultmag = 0;

  // generate filters and compute max and min values at the same time

  for(int x=0;x<size;x++)
  {
    xoff = x - floor(size/2);
    for(int y=0;y<size;y++)
    {
      yoff = y - floor(size/2);
      rx = ((xoff-x0)*cos(theta))+((yoff-y0)*sin(theta));
      ry = (-(xoff-x0)*sin(theta))+((yoff-y0)*cos(theta));
      re = cos(2*M_PI*((u0*xoff)+(v0*yoff))+P)*(K*exp(-M_PI*((pow(a,2)*pow(rx,2))+(pow(b,2)*pow(ry,2)))));
      im = sin(2*M_PI*((u0*xoff)+(v0*yoff))+P)*(K*exp(-M_PI*((pow(a,2)*pow(rx,2))+(pow(b,2)*pow(ry,2)))));
      filter[x][y] = complex<double>(re,im);
      if(filter[x][y].real() > maxfilterreal) maxfilterreal = filter[x][y].real();
      else if(filter[x][y].real() < minfilterreal) minfilterreal = filter[x][y].real();
      if(filter[x][y].imag() > maxfilterimag) maxfilterimag = filter[x][y].imag();
      else if(filter[x][y].imag() < minfilterimag) minfilterimag = filter[x][y].imag();
      if(abs(filter[x][y]) > maxfiltermag) maxfiltermag = abs(filter[x][y]);
      else if(abs(filter[x][y]) < minfiltermag) minfiltermag = abs(filter[x][y]);
    }
  }

  // create and initialize zeropadded matrix

  procimg = new double*[origimage->width()+size+1];
  for(int i=0;i<origimage->width()+size+1;i++)
    procimg[i] = new double[origimage->height()+size+1];
  for(int x=0;x<origimage->width()+size+1;x++)
    for(int y=0;y<origimage->height()+size+1;y++)
      procimg[x][y] = 0;

  // copy original image into a zeropadded matrix and create result matrix at the same time

  result = new complex<double>*[origimage->width()];
  for(int x=0;x<origimage->width();x++)
  {
    result[x] = new complex<double>[origimage->height()];
    for(int y=0;y<origimage->height();y++)
    {
      temprgb = origimage->pixel(x,y);
      procimg[x+int(ceil(size/2))][y+int(ceil(size/2))] = double(qGray(temprgb));
    }
  }

  // create progressbar and display this for the convolution process

  delete progress;
  progress = new QProgressDialog("Filtering image...", "Cancel", origimage->width(), this, "progress", TRUE);

  // perform convolution, store max and min values at the same time

  for(int x=0;x<origimage->width();x++)
  {
    progress->setProgress(x);
    if(progress->wasCancelled())
      break;    
    for(int y=0;y<origimage->height();y++)
    {
      result[x][y]=0;
      for(int xx=0;xx<size;xx++)
	for(int yy=0;yy<size;yy++)
	  result[x][y] += procimg[x+xx][y+yy]*filter[size-1-xx][size-1-yy];
      if(result[x][y].real() > maxresultreal) maxresultreal = result[x][y].real();
      else if(result[x][y].real() < minresultreal) minresultreal = result[x][y].real();
      if(result[x][y].imag() > maxresultimag) maxresultimag = result[x][y].imag();
      else if(result[x][y].imag() < minresultimag) minresultimag = result[x][y].imag();
      if(abs(result[x][y]) > maxresultmag) maxresultmag = abs(result[x][y]);
      else if(abs(result[x][y]) < minresultmag) minresultmag = abs(result[x][y]);
    }
  }
  progress->setProgress(origimage->width());

  // update images for viewing

  filterreal->setImage(&*filter, size, size, maxfilterreal, minfilterreal, 0);
  filterimag->setImage(&*filter, size, size, maxfilterimag, minfilterimag, 1);
  filtermag->setImage(&*filter, size, size, maxfiltermag, minfiltermag, 2);
  resultreal->setImage(&*result, origimage->width(), origimage->height(), maxresultreal, minresultreal, 0);
  resultimag->setImage(&*result, origimage->width(), origimage->height(), maxresultimag, minresultimag, 1);
  resultmag->setImage(&*result, origimage->width(), origimage->height(), maxresultmag, minresultmag, 2);

  // display magnitude result

  viewimage = resultmag;
  repaint();

}

