#include <cmath>
#include <complex>
#include <iostream>
#include <qpainter.h>
#include "grayprocessingimage.h"
#include "commongrayproc.h"

// C O N S T R U C T O R S

//! Constructor taking no argument. Converts the null image to grayscale.
GrayProcessingImage::GrayProcessingImage()
  : QImage()
{
  toGrayscale();
}

//! Copy constructor. Send the image to QImage.
GrayProcessingImage::GrayProcessingImage(const GrayProcessingImage& image)
  : QImage(image)
{
}

//! Constructor taking a QImage as argument. Converts the image to grayscale.
GrayProcessingImage::GrayProcessingImage(QImage image)
  : QImage(image)
{
  toGrayscale();
}

//! Constructor taking size, depth etc. as arguments. Passes to QImage.
GrayProcessingImage::GrayProcessingImage(const QSize& size, int depth = 8,
					 int numColors=256,
					 Endian bitOrder=IgnoreEndian)
  : QImage(size, depth, numColors, bitOrder)
{
  toGrayscale();
}

//! Constructor taking size, depth etc. as arguments. Passes to QImage.
GrayProcessingImage::GrayProcessingImage(int width, int height, int depth=8,
					 int numColors=256,
					 Endian bitOrder=IgnoreEndian)
  : QImage(width, height, depth, numColors, bitOrder)
{
  toGrayscale();
}

// P R I V A T E   F U N C T I O N S

//! Fill the palette with 256 gray values.
void GrayProcessingImage::fillGrayPalette()
{
  setNumColors(256);
  for (int i = 0; i < 256; i++)
    setColor(i, qRgb(i, i, i));
}

//! Bilinear interpolation of pixel values
uchar GrayProcessingImage::interpolate(double x, double y)
{
  int w = width(), h = height();
  if ((x < 0) || (y < 0) || (int(x) > w - 2) || int(y) > h - 2) {
    return 0;
  }
  uchar *bts = bits();
  int i = int(x), j = int(y);
  double alpha = x - i, beta = y - j;
  uchar pij = *(bts + i + w*j);
  uchar pi1j = *(bts + i + 1 + w*j);
  uchar pij1 = *(bts + i + w*(j + 1));
  uchar pi1j1 = *(bts + i + 1 + w*(j + 1));
  return toChar((1 - alpha)*(1 - beta)*pij +
		alpha*(1 - beta)*pi1j +
		(1 - alpha)*beta*pij1 +
		alpha*beta*pi1j1);
}

//! Converts the QImage in any format to a similar 8 bpp grayscale image.
/*! Currently, only 8 bpp and 32 bpp images are supported. */
void GrayProcessingImage::toGrayscale()
{
  QImage oldimage(copy());
  create(oldimage.size(), 8, 256);
  long w = width(), h = height();
  fillGrayPalette();
  switch(oldimage.depth()) {
  case 8:
    {
      uchar *oldbits = oldimage.bits();
      uchar *newbits = bits();
      QRgb *colortable = oldimage.colorTable();
      for (long i = 0; i < w*h; i++)
	*newbits++ = qGray(colortable[*oldbits++]);
    }
    break;
  case 16:
    {
      // Not implemented yet.
    }
  case 32:
    {
      for (int i = 0; i < w; i++)
	for (int j = 0; j < h; j++)
	  setPixel(i, j, qGray(oldimage.pixel(i, j)));
    }
    break;
  }
}

//! Calculate coefficients for intensity plane.
void GrayProcessingImage::calculateIntensityPlane(double *a,
						  double *b, double *c)
{
  uchar *newbits = bits();
  int h = height(), w = width();
  // Determine the coefficients of the least square matrix:
  double A = 0, B = 0, C = 0,
    D = 0, E = 0, F = 0,
    G = 0, H = 0, I = 0,
    P = 0, Q = 0, R = 0;
  for (int i = 0; i < w; i++)
    for (int j = 0; j < h; j++) {
      A += i*i; B += i*j; C += i;
      E += j*j; F += j; I++;
      P += double(i)*(*(newbits + i + w*j));
      Q += double(j)*(*(newbits + i + w*j));
      R += *(newbits + i + w*j);
    }
  D = B; G = C; H = F;
  // Determine the parameters of the plane from the coefficients
  *c = (Q - P*D/A - (R - Q*G/D)*(E - B*D/A)/(H - E*G/D))/
    (F - C*D/A - (I - F*G/D)*(E - B*D/A)/(H - E*G/D));
  *b = (Q - P*D/A - *c*(F - C*D/A))/(E - B*D/A);
  *a = (P - *b*B - *c*C)/A;
}

//! Generate histogram of image.
void GrayProcessingImage::calculateHistogram(unsigned long *histogram)
{
  unsigned long i;
  int w = width(), h = height();
  // Reset histogram
  for (i = 0; i < 256; i++)
    histogram[i] = 0;
  uchar *newbits = bits();
  // Calculate histogram
  unsigned long numpixels = w*h;
  for (i = 0; i < numpixels; i++)
    histogram[*newbits++]++;
}

// P U B L I C   F U N C T I O N S

//! Load image from file, convert to grayscale.
bool GrayProcessingImage::load(const QString& fileName, const char *format=0)
{
  bool ret = QImage::load(fileName, format);
  toGrayscale();
  return ret;
}

//! Resize the image using bilinear interpolation.
/*! For enlargements or small reductions this is better than the
    function QImage::smoothScale(), Copyright (C) 1989, 1991 by Jef
    Poskanzer. */
GrayProcessingImage GrayProcessingImage::bilinearResize(int neww, int newh)
{
  GrayProcessingImage newimage(neww, newh);
  int oldh = height(), oldw = width();
  double hscale = double(oldh - 1)/(newh - 1);
  double wscale = double(oldw - 1)/(neww - 1);
  double oldi, oldj;
  //  uchar *newbits = newimage.bits();
  for (int j = 0; j < newh; j++)
    for (int i = 0; i < neww; i++) {
      oldi = wscale*i; oldj = hscale*j;
      //      *newbits++ = interpolate(oldi, oldj);
      newimage.setPixel(i, j, interpolate(oldi, oldj));
    }
  return newimage;
}

//! Convert image to its own intensity plane.
GrayProcessingImage GrayProcessingImage::intensityPlane()
{
  double a, b, c;
  calculateIntensityPlane(&a, &b, &c);
  GrayProcessingImage newimage(size());
  newimage.fillGrayPalette();
  uchar *newbits = newimage.bits();
  int w = width(), h = height();
  for (int j = 0; j < h; j++)
    for (int i = 0; i < w; i++)
      *newbits++ = toChar(a*i + b*j + c);
  return newimage;
}

//! Subtract the image's intensity plane.
GrayProcessingImage GrayProcessingImage::subtractIntensityPlane()
{
  GrayProcessingImage newimage(size());
  double a, b, c;
  calculateIntensityPlane(&a, &b, &c);
  uchar *oldbits = bits();
  uchar *newbits = newimage.bits();
  int w = width(), h = height();
  for (int j = 0; j < h; j++)
    for (int i = 0; i < w; i++)
      *newbits++ = toChar(*oldbits++ - toChar(a*i + b*j + c) + 128);
  return newimage;
}

//! Perform histogram equalization.
GrayProcessingImage GrayProcessingImage::histogramEqualize()
{
  int w = width(), h =height();
  unsigned long i;
  unsigned long histogram[256];
  unsigned long sumhist[256];
  calculateHistogram(histogram);
  //! Calculate normalized sum of histogram
  unsigned long sum = 0;
  unsigned long numpixels = w*h;
  double scalefactor = 255.0 / numpixels;
  for (i = 0; i < 256; i++) {
    sum += histogram[i];
    sumhist[i] = (unsigned long)((sum*scalefactor) + 0.5);
  }
  //! Transform image using sumhist as LUT
  GrayProcessingImage newimage(size());
  uchar *oldbits = bits();
  uchar *newbits = newimage.bits();
  for (i = 0; i < numpixels; i++)
    *newbits++ = sumhist[*oldbits++];
  return newimage;
}

//! Applies 3x3 convolution given by matrix on image.
/*! Last parameter indicate type of padding; one of the following:
    GrayProcessingImage::zeroPadding, GrayProcessingImage::copyPadding
    or GrayProcessingImage::wrapPadding. Only zeropadding is
    implemented so far. */
GrayProcessingImage GrayProcessingImage::convolution3x3(double matrix[3][3],
							int add = 0,
							PaddingType pad 
							= zeroPadding)
{
  GrayProcessingImage newimage(size());
  uchar *newbits = newimage.bits();
  uchar *oldbits = bits();
  int w = width(), h = height();
  for (int j = 1; j < h - 1; j++)
    for (int i = 1; i < w - 1; i++)
      *(newbits + i + w*j) =
	toChar(matrix[0][0]*(*(oldbits + i - 1 + w*(j - 1))) +
	       matrix[0][1]*(*(oldbits + i + w*(j - 1))) +
	       matrix[0][2]*(*(oldbits + i + 1 + w*(j - 1))) +
	       matrix[1][0]*(*(oldbits + i - 1 + w*j)) +
	       matrix[1][1]*(*(oldbits + i + w*j)) +
	       matrix[1][2]*(*(oldbits + i + 1 + w*j)) +
	       matrix[2][0]*(*(oldbits + i - 1 + w*(j + 1))) +
	       matrix[2][1]*(*(oldbits + i + w*(j + 1))) +
	       matrix[2][2]*(*(oldbits + i + 1 + w*(j + 1))) + add);
  switch (pad) {
  case zeroPadding:
    {
      for (int i = 1; i < w - 1; i++) {
	// Upper line (j = 0):
	*(newbits + i) = 
	  toChar(matrix[1][0]*(*(oldbits + i - 1)) +
		 matrix[1][1]*(*(oldbits + i)) +
		 matrix[1][2]*(*(oldbits + i + 1)) +
		 matrix[2][0]*(*(oldbits + i - 1 + w)) +
		 matrix[2][1]*(*(oldbits + i + w)) +
		 matrix[2][2]*(*(oldbits + i + 1 + w)) + add);
	// Lower line (j = h - 1):
	*(newbits + i + w*(h - 1)) =
	  toChar(matrix[0][0]*(*(oldbits + i - 1 + w*(h - 2))) +
		 matrix[0][1]*(*(oldbits + i + w*(h - 2))) +
		 matrix[0][2]*(*(oldbits + i + 1 + w*(h - 2))) +
		 matrix[1][0]*(*(oldbits + i - 1 + w*(h - 1))) +
		 matrix[1][1]*(*(oldbits + i + w*(h - 1))) +
		 matrix[1][2]*(*(oldbits + i + 1 + w*(h - 1))) + add);
      }
      for (int j = 1; j < h - 1; j++) {
	// Leftmost line (i = 0):
	*(newbits + w*j) =
	  toChar(matrix[0][1]*(*(oldbits + w*(j - 1))) +
		 matrix[0][2]*(*(oldbits + 1 + w*(j - 1))) +
		 matrix[1][1]*(*(oldbits + w*j)) +
		 matrix[1][2]*(*(oldbits + 1 + w*j)) +
		 matrix[2][1]*(*(oldbits + w*(j + 1))) +
		 matrix[2][2]*(*(oldbits + 1 + w*(j + 1))) + add);
	// Rightmost line (i = w - 1):
	*(newbits + w - 1 + w*j) =
	  toChar(matrix[0][0]*(*(oldbits + w - 2 + w*(j - 1))) +
		 matrix[0][1]*(*(oldbits + w - 1 + w*(j - 1))) +
		 matrix[1][0]*(*(oldbits + w - 2 + w*j)) +
		 matrix[1][1]*(*(oldbits + w - 1 + w*j)) +
		 matrix[2][0]*(*(oldbits + w - 2 + w*(j + 1))) +
		 matrix[2][1]*(*(oldbits + w - 1+ w*(j + 1))) + add);
      }
      // Corners:
      *newbits =
	toChar(matrix[2][2]*(*oldbits) + add);
      *(newbits + w - 1) =
	toChar(matrix[2][0]*(*(oldbits + w - 1)) + add);
      *(newbits + w*(h - 1)) =
	toChar(matrix[0][2]*(*(oldbits + w*(h - 1))) + add);
      *(newbits + w*(h - 1) + w - 1) =
	toChar(matrix[2][2]*(*(oldbits + w*(h - 1) + w - 1)) + add);
    }
    break;
  case copyPadding:
    {
      // Not implemented yet.
    }
    break;
  case wrapPadding:
    {
      // Not implemented yet.
    }
    break;
  }
  return newimage;
}

//! Make the average of the given images a new image.
GrayProcessingImage GrayProcessingImage::average(vector<GrayProcessingImage>
						 images)
{
  GrayProcessingImage newimage(images[0].size());
  newimage.detach();
  int h = height(), w = width(), numimages = images.size();
  for (int i = 0; i < w; i++)
    for (int j = 0; j < h; j++) {
      long bit = 0;
      for (int n = 0; n < numimages; n++)
	bit += images[n].pixelIndex(i, j);
      newimage.setPixel(i, j, bit/numimages);
    }
  return newimage;
}

//! Normalize (geometrically) image of face with given eye coordinates.
/*! Normalization involves rotating the image so that the eyes are
    located on an horizontal line, scaling the image so that the
    distance between the eyes are fixed, and making a new image of
    given size and position of the face. The other parameters mean:
    pctabove - the amount of the image above the eyeline (in pct),
    pctofwidth - the ratio of distance between eyes to the width of
    the image (in pct), neww & newh - the dimensions of the new
    image. */
GrayProcessingImage
GrayProcessingImage::faceNormalize(int lx, int ly, int rx, int ry,
				   int pctabove = 30, int pctofwidth = 60,
				   int neww = 20, int newh = 20)
{
  GrayProcessingImage newimage(neww, newh);
  // Old coordinates of center of eyes
  int oldcx = (lx + rx)/2, oldcy = (ly + ry)/2;
  // New coordinates of center of eyes
  int newcx = neww/2, newcy = (newh*pctabove)/100;
  // Old distance between eyes
  double olddist = sqrt((rx - lx)*(rx - lx) + (ry - ly)*(ry - ly));
  // New distance between eyes, and scaling needed
  double newdist = (double(neww)*pctofwidth)/100;
  double scale = olddist/newdist;
  // The neede rotation angle 
  double theta = -atan(float(ry - ly)/(rx - lx));
  double costh = cos(theta), sinth = sin(theta), tanth = tan(theta);
  // Real coordinates of the original image (this)
  double x, y, unscaley;
  // Perform combined scale/rotate operation:
  uchar *newbits = newimage.bits();
  for (int j = 0; j < newh; j++)
    for (int i = 0; i < neww; i++) {
      // Determine corresponding x and y from original image
      unscaley = ((j - newcy) - tanth*(i - newcx)) /
  	(sinth*tanth + costh);
      x = (scale*((i - newcx) + sinth*unscaley))/costh + oldcx;
      y = scale*unscaley + oldcy;
      *newbits++ = interpolate(x, y);
    }
  return newimage;
}


//! Calculate first component of image normal.
GrayProcessingImage GrayProcessingImage::normal1()
{
  GrayProcessingImage newimage = copy();
  int w = width(), h = height();
  for (int i = 0; i < w - 1; i++)
    for (int j = 0; j < h - 1; j++)
      newimage.setPixel(i, j,
			(-pixelIndex(i + 1, j + 1) + pixelIndex(i, j + 1) +
			 pixelIndex(i, j) - pixelIndex(i + 1, j))/4
			       + 128);
  return newimage;
}

//! Calculate second component of image normal.
GrayProcessingImage GrayProcessingImage::normal2()
{
  GrayProcessingImage newimage = copy();
  int w = width(), h = height();
  for (int i = 0; i < w - 1; i++)
    for (int j = 0; j < h - 1; j++)
      newimage.setPixel(i, j,
			(-pixelIndex(i + 1, j + 1) - pixelIndex(i, j + 1) +
			 pixelIndex(i, j) + pixelIndex(i + 1, j))/4
			       + 128);
  return newimage;
}

//! Returns two images describing the directional image.
/*! The parameters windowSize gives the size of the NxN average
    frame. The first of the returned images is the direction (0: -90,
    255: +90), and the second is a weighted sum of the contrast and
    the consistency. */
pair<GrayProcessingImage, GrayProcessingImage>
GrayProcessingImage::dirImage(int windowSize)
{
  int w = width(), h = height();
  int nw = w/windowSize, nh = h/windowSize;
  GrayProcessingImage norm1 = normal1();
  GrayProcessingImage norm2 = normal2();
  GrayProcessingImage dir(nw, nh);
  GrayProcessingImage weight(nw, nh);
  for (int i = 0; i < nw - 1; i++)
    for (int j = 0; j < nh - 1; j++) {
      int oi = i*windowSize;
      int oj = j*windowSize;
      double ang;
      double nx = 0, ny = 0;
      for (int wi = 0; wi < windowSize; wi++)
	for (int wj = 0; wj < windowSize; wj++) {
	  nx += norm1.pixelIndex(oi + wi, oj + wj) - 128;
	  ny += norm2.pixelIndex(oi + wi, oj + wj) - 128;
	}
        nx /= windowSize;
        ny /= windowSize;
      double nnorm = sqrt(nx*nx + ny*ny);
      weight.setPixel(i, j, toChar(nnorm*8));
      ang = atan2(ny, nx) + M_PI/2;
      while (ang > M_PI/2)
	ang -= M_PI;
      dir.setPixel(i, j, toChar((ang + M_PI/2)*255/M_PI));
// Old code using algorithm from paper:
//
//        double A = 0, B = 0, C = 0;
//        for (int wi = 0; wi < windowSize; wi++)
//  	for (int wj = 0; wj < windowSize; wj++) {
//  	  double v1 = double(norm1.pixelIndex(oi + wi, oj + wj)) - 128;
//  	  double v2 = double(norm2.pixelIndex(oi + wi, oj + wj)) - 128;
//  	  A += v1*v1;
//  	  B += v2*v2;
//  	  C += v1*v2;
//  	}
//        if (C == 0) {
//  	if (A < B) {
//  	  u1 = 1; u2 = 0;
//  	  ang = M_PI/2;
//  	}
//  	else {
//  	  u1 = 0; u2 = -1;
//  	  ang = 0;
//  	}
//        }
//        else {
//  	//	u2 = (B - A)/(2*C) - sqrt(((B - A)/(2*C))*((B - A)/(2*C)) + 1);
//  	u2 = (B - A - sqrt((A - B)*(A - B) + 4*C*C))/2;
//  	double norm = sqrt(1+u2*u2);
//  	u1 = 1.0/norm;
//  	u2 *= u1;
//  	ang = atan2(u2, u1);
//  	if (ang > M_PI/2)
//  	  ang -= M_PI;
//  	if (ang < -M_PI/2)
//  	  ang += M_PI;
//        }
//        dir.setPixel(i, j, toChar((ang + M_PI/2)*255/M_PI));
//        double Cnorm = (A + B)/(128*128*windowSize*windowSize);
//        double Enorm = (A + B - sqrt((A - B)*(A - B) + 4*C*C))/(A + B);
//        weight.setPixel(i, j, toChar((.2*Cnorm + .8*(1-Enorm))*255));
//        dir.setPixel(i, j, toChar((u1 + 1)*128));
//        weight.setPixel(i, j, toChar((u2 + 1)*128));
    }
  return pair<GrayProcessingImage, GrayProcessingImage>(dir, weight);
}

//! Creates an image that visualises the directional image.
GrayProcessingImage
GrayProcessingImage::visualiseDirImage(pair<GrayProcessingImage,
				       GrayProcessingImage> dir)
{
  int dw = dir.first.width(), dh = dir.first.height();
  QPixmap newpixmap(dw*10, dh*10);
  newpixmap.fill();
  QPainter p(&newpixmap);
  for (int i = 0; i < dw; i++)
    for (int j = 0; j < dh; j++) {
      double ang = double(dir.first.pixelIndex(i, j))/255*M_PI - M_PI/2;
      double length = double(dir.second.pixelIndex(i, j))/255;
      int xl = int(length*cos(ang)*10);
      int yl = int(length*sin(ang)*10);
      p.drawLine(i*10, j*10, i*10 + xl, j*10 + yl);
    }
  return GrayProcessingImage(newpixmap.convertToImage());
}  

//! Creates an image that visualises the directional image.
GrayProcessingImage GrayProcessingImage::visualiseDirImage(int windowSize)
{
  return visualiseDirImage(dirImage(windowSize));
}

//! Return the average pixel value in the image.
double GrayProcessingImage::average()
{
  int n = height()*width();
  uchar *bts = bits();
  double sum = 0;
  for (int i = 0; i < n; i++)
    sum += *bts++;
  return sum/n;
}

//! Return the variance of the pixel values in the image.
double GrayProcessingImage::variance()
{
  int n = height()*width();
  double avg = average();
  double var = 0;
  uchar *bts = bits();
  for (int i = 0; i < n; i++) {
    var += (*bts - avg)*(*bts - avg);
    bts++;
  }
  return var;
}

//! Return the correlation coefficient between two images.
/*! As defined in Gonzales & Woods p. 584, eq. 9.3-8. Assumes that the
    two images have the same dimensions; if not, the result is
    undefined. In order to make the correlation act as a distance, it
    is inverted and scaled such that correlation 1 corresponds to a
    return value '0' (no distance between images, and correlation -1
    corresponds to 1 ('100% distance').*/
double GrayProcessingImage::correlcoeff(GrayProcessingImage image)
{
  int n = height()*width();
  double avg = average(),
    imavg = image.average(),
    var = variance(),
    imvar = image.variance();
  if (var == 0 || imvar == 0)
    return 1;
  uchar *bts = bits();
  uchar *imbts = image.bits();
  double cov = 0;
  for (int i = 0; i < n; i++)
    cov += (*bts++ - avg)*(*imbts++ - imavg);
  return 0.5 - cov/(2*sqrt(var*imvar));
}

//! Changes the content of the image.
/*! From a complex double array, sets the content of the image to either the real, imaginary or magnitude of the array.*/
void GrayProcessingImage::setImage(complex<double> **img, int w, int h, double max, double min, int i)
{
  reset();
  create(w,h,8,256);
  fillGrayPalette();

  if(i == 0)
    if(min < 0)
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor(((img[x][y].real()+fabs(min))/(max+fabs(min)))*255)));
    else
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor((img[x][y].real()/max)*255)));
  else if(i == 1)
    if(min < 0)
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor(((img[x][y].imag()+fabs(min))/(max+fabs(min)))*255)));
    else
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor((img[x][y].imag()/max)*255)));
  else
    if(min < 0)
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor(((abs(img[x][y])+fabs(min))/(max+fabs(min)))*255)));
    else
      for(int x=0;x<w;x++)
	for(int y=0;y<h;y++)
	  setPixel(x,y,int(floor((abs(img[x][y])/max)*255)));
}



